from typing import Optional, Tuple
import re
from PySide6.QtGui import QColor, QSyntaxHighlighter, QTextDocument

NO_STATUS = 0
STATUS_GOOD = 1
STATUS_BAD = 2

# palette from: https://github.com/catppuccin/catppuccin
FG =      '#cad3f5'
BG =      '#24273a'
SEL =     '#363a4f'
KEYWORD = '#8aadf4'
IDENT =   '#91d7e3'
NUM =     '#ed8796'
OP =      '#eed49f'
BG_GOOD = '#36504f'
BG_BAD  = '#863a4f'

class ChypHighlighter(QSyntaxHighlighter):
    def __init__(self, doc: QTextDocument) -> None:
        super().__init__(doc)
        self.region: Optional[Tuple[int,int]] = None
        self.region_status = NO_STATUS

    def set_current_region(self, region: Optional[Tuple[int,int]], status: int) -> None:
        self.region = region
        self.region_status = status
        self.rehighlight()

    def highlightBlock(self, text: str) -> None:
        ident = '[a-zA-Z_][a-zA-Z0-9_]*'
        for m in re.finditer('(^|\\W)(let|gen|rule|by|rewrite)\\s+(%s)?' % ident, text):
            x,y = m.span(2)
            self.setFormat(x, y-x, QColor(KEYWORD))
            x,y = m.span(3)
            self.setFormat(x, y-x, QColor(IDENT))

        for m in re.finditer('[?.>:;*=\\-\\[\\]]', text):
            self.setFormat(m.start(), 1, QColor(OP))

        for m in re.finditer('(\\W|^)([0-9]+)(\\W|$)', text):
            x,y = m.span(2)
            self.setFormat(x, y-x, QColor(NUM))

        # highlight the region that is currently in focus
        if self.region:
            x, y = self.region
            start = self.currentBlock().position()
            length = self.currentBlock().length()
            end = start + length
            if y >= start and x < end:
                x_rel = max(x - start, 0)
                y_rel = min(y - start, length)
                for c in range(x_rel, y_rel):
                    f = self.format(c)

                    if self.region_status == STATUS_GOOD:
                        f.setBackground(QColor(BG_GOOD))
                    elif self.region_status == STATUS_BAD:
                        f.setBackground(QColor(BG_BAD))
                    else:
                        f.setBackground(QColor(SEL))
                    self.setFormat(c, 1, f)




#     chyp - A compositional hypergraph library
#     Copyright (C) 2022 - Aleks Kissinger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from PySide6.QtCore import QByteArray, QSettings
from PySide6.QtGui import QCloseEvent
from PySide6.QtWidgets import QMainWindow, QVBoxLayout, QWidget

from .graphview import GraphView

class MainWindow(QMainWindow):
    def __init__(self) -> None:
        super().__init__()
        conf = QSettings('chyp', 'chyp')

        self.setWindowTitle("chyp")

        w = QWidget(self)
        w.setLayout(QVBoxLayout())
        self.setCentralWidget(w)
        w.layout().setContentsMargins(0,0,0,0)
        w.layout().setSpacing(0)
        self.resize(1600, 800)
        
        geom = conf.value("main_window_geometry")
        if geom and isinstance(geom, QByteArray): self.restoreGeometry(geom)
        self.show()

        self.view = GraphView()
        # self.view.set_graph(g)
        w.layout().addWidget(self.view)



    def closeEvent(self, e: QCloseEvent) -> None:
        conf = QSettings('chyp', 'chyp')
        conf.setValue("main_window_geometry", self.saveGeometry())
        e.accept()



from typing import Optional, Tuple

from PySide6.QtGui import QTextCursor
from PySide6.QtWidgets import QPlainTextEdit

from .document import ChypDocument
from .highlighter import BG, FG, NO_STATUS


class CodeView(QPlainTextEdit):
    def __init__(self) -> None:
        super().__init__()
        self.setStyleSheet("QPlainTextEdit {background-color: %s; color: %s}" % (BG, FG))

    def set_current_region(self, region: Optional[Tuple[int,int]], status: int = NO_STATUS) -> None:
        doc = self.document()

        if isinstance(doc, ChypDocument):
            self.blockSignals(True)
            doc.highlighter.set_current_region(region, status)
            self.blockSignals(False)

    def add_line_below(self, text: str) -> None:
        doc = self.document()

        if isinstance(doc, ChypDocument):
            cursor = self.textCursor()
            if doc.highlighter.region:
                cursor.setPosition(doc.highlighter.region[1])
            cursor.movePosition(QTextCursor.MoveOperation.EndOfLine)
            cursor.insertText('\n' + text)
            self.setTextCursor(cursor)





#     chyp - A compositional hypergraph library
#     Copyright (C) 2022 - Aleks Kissinger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Callable, Dict, Optional, Tuple
from PySide6.QtCore import QByteArray, QFileInfo, QObject, QThread, Qt, QSettings
from PySide6.QtGui import QCloseEvent, QKeySequence, QTextCursor
from PySide6.QtWidgets import QApplication, QHBoxLayout, QMainWindow, QMenuBar, QSplitter, QVBoxLayout, QWidget

from chyp.term import graph_to_term

from ..layout import convex_layout
from ..graph import Graph
from ..state import RewriteState, State
from ..matcher import match_rule
from ..rewrite import rewrite

from .graphview import GraphView
from .codeview import CodeView
from .document import ChypDocument
from .highlighter import STATUS_GOOD, STATUS_BAD

class Editor(QMainWindow):
    def __init__(self) -> None:
        super().__init__()
        conf = QSettings('chyp', 'chyp')

        self.setWindowTitle("chyp")

        w = QWidget(self)
        w.setLayout(QVBoxLayout())
        self.setCentralWidget(w)
        w.layout().setContentsMargins(0,0,0,0)
        w.layout().setSpacing(0)
        self.resize(1600, 800)
        
        geom = conf.value("editor_window_geometry")
        if geom and isinstance(geom, QByteArray): self.restoreGeometry(geom)
        self.show()

        # save splitter position
        self.splitter = QSplitter(Qt.Orientation.Vertical)
        w.layout().addWidget(self.splitter)

        self.state = State()

        self.lhs_view = GraphView()
        self.rhs_view = GraphView()
        self.rhs_view.setVisible(False)

        self.graph_panel = QWidget(self)
        self.graph_panel.setLayout(QHBoxLayout())
        self.graph_panel.layout().addWidget(self.lhs_view)
        self.graph_panel.layout().addWidget(self.rhs_view)
        self.splitter.addWidget(self.graph_panel)

        self.code_view = CodeView()
        self.doc = ChypDocument(self)
        self.code_view.setDocument(self.doc)
        self.doc.fileNameChanged.connect(self.update_file_name)
        self.doc.modificationChanged.connect(self.update_file_name)
        self.doc.recentFilesChanged.connect(self.update_recent_files)
        self.update_file_name()

        self.splitter.addWidget(self.code_view)
        self.code_view.setFocus()

        self.build_menu()

        splitter_state = conf.value("editor_splitter_state")
        if splitter_state and isinstance(splitter_state, QByteArray): self.splitter.restoreState(splitter_state)

        self.code_view.cursorPositionChanged.connect(self.show_at_cursor)
        self.code_view.textChanged.connect(self.invalidate_text)
        self.parsed = True

        # keep a cache of graphs that have already been laid out
        self.graph_cache : Dict[int, Tuple[Graph, Optional[Graph]]] = dict()

    def build_menu(self) -> None:
        menu = QMenuBar()
        file_menu = menu.addMenu("&File")
        edit_menu = menu.addMenu("&Edit")
        code_menu = menu.addMenu("&Code")

        file_new = file_menu.addAction("&New")
        file_new.triggered.connect(self.doc.new)

        file_open = file_menu.addAction("&Open")
        file_open.setShortcut(QKeySequence(QKeySequence.StandardKey.Open))
        file_open.triggered.connect(lambda: self.doc.open())

        self.file_open_recent = file_menu.addMenu("Open &Recent")
        self.update_recent_files()

        file_menu.addSeparator()

        file_save = file_menu.addAction("&Save")
        file_save.setShortcut(QKeySequence(QKeySequence.StandardKey.Save))
        file_save.triggered.connect(self.doc.save)

        file_save_as = file_menu.addAction("Save &As")
        file_save_as.setShortcut(QKeySequence(QKeySequence.StandardKey.SaveAs))
        file_save_as.triggered.connect(self.doc.save_as)

        file_menu.addSeparator()

        file_exit = file_menu.addAction("E&xit")
        file_exit.setShortcut(QKeySequence(QKeySequence.StandardKey.Quit))

        app = QApplication.instance()
        if app:
            file_exit.triggered.connect(app.quit)

        edit_undo = edit_menu.addAction("&Undo")
        edit_undo.setShortcut(QKeySequence(QKeySequence.StandardKey.Undo))
        edit_undo.triggered.connect(self.code_view.undo)

        edit_redo = edit_menu.addAction("&Redo")
        edit_redo.setShortcut(QKeySequence(QKeySequence.StandardKey.Redo))
        edit_redo.triggered.connect(self.code_view.redo)

        code_run = code_menu.addAction("&Run")
        code_run.setShortcut(QKeySequence("Ctrl+R"))
        code_run.triggered.connect(self.update)

        code_add_rewrite_step = code_menu.addAction("&Add Rewrite Step")
        code_add_rewrite_step.setShortcut(QKeySequence("Ctrl+Return"))
        code_add_rewrite_step.triggered.connect(lambda: self.code_view.add_line_below("  = ? by "))

        code_add_rewrite_step = code_menu.addAction("&Repeat Rewrite Step")
        code_add_rewrite_step.setShortcut(QKeySequence("Ctrl+Shift+Return"))
        code_add_rewrite_step.triggered.connect(self.repeat_step_at_cursor)

        code_next_rewrite = code_menu.addAction("&Next Rewrite")
        code_next_rewrite.setShortcut(QKeySequence("Ctrl+N"))
        code_next_rewrite.triggered.connect(self.next_rewrite_at_cursor)

        code_menu.addSeparator()

        code_next_part = code_menu.addAction("Next &Part")
        code_next_part.setShortcut(QKeySequence("Ctrl+J"))
        code_next_part.triggered.connect(lambda: self.next_part(step=1))

        code_previous_part = code_menu.addAction("Previous &Part")
        code_previous_part.setShortcut(QKeySequence("Ctrl+K"))
        code_previous_part.triggered.connect(lambda: self.next_part(step=-1))

        self.setMenuBar(menu)

    def update_file_name(self) -> None:
        title = 'chyp - '
        if self.doc.file_name:
            fi = QFileInfo(self.doc.file_name)
            title += fi.fileName()
        else:
            title += 'Untitled'

        if self.doc.isModified():
            title += '*'

        self.setWindowTitle(title)

    def update_recent_files(self) -> None:
        def open_recent(f: str) -> Callable:
            return lambda: self.doc.open(f)

        self.file_open_recent.clear()
        for f in self.doc.recent_files():
            fi = QFileInfo(f)
            action = self.file_open_recent.addAction(fi.fileName())
            action.triggered.connect(open_recent(f))

    def invalidate_text(self) -> None:
        self.parsed = False
        self.graph_cache = dict()
        self.code_view.set_current_region(None)
        self.update_state(quiet=True)

    def next_part(self, step:int=1) -> None:
        if not self.parsed: return

        cursor = self.code_view.textCursor()
        pos = cursor.position()
        p = self.state.part_with_index_at(pos)
        i = p[0] if p else 0
        i += step

        if i >= 0 and i < len(self.state.parts):
            p1 = self.state.parts[i]
            cursor.setPosition(p1[1])
            self.code_view.setTextCursor(cursor)
            

    def show_at_cursor(self) -> None:
        if not self.parsed: return

        pos = self.code_view.textCursor().position()
        p = self.state.part_with_index_at(pos)
        if p:
            i, part = p
            self.code_view.set_current_region((part[0], part[1]))
            if part[2] in ('let','gen') and part[3] in self.state.graphs:
                if i not in self.graph_cache:
                    g = self.state.graphs[part[3]].copy()
                    convex_layout(g)
                    self.graph_cache[i] = (g, None)
                else:
                    g, _ = self.graph_cache[i]
                self.rhs_view.setVisible(False)
                self.lhs_view.set_graph(g)
            elif part[2] == 'rule' and part[3] in self.state.rules:
                if i not in self.graph_cache:
                    lhs = self.state.rules[part[3]].lhs.copy()
                    rhs = self.state.rules[part[3]].rhs.copy()
                    convex_layout(lhs)
                    convex_layout(rhs)
                    self.graph_cache[i] = (lhs, rhs)
                else:
                    lhs, rhs0 = self.graph_cache[i]
                    if not rhs0: raise ValueError("Rule in graph_cache should have RHS")
                    rhs = rhs0
                self.rhs_view.setVisible(True)
                self.lhs_view.set_graph(lhs)
                self.rhs_view.set_graph(rhs)
            elif part[2] == 'rewrite' and part[3] in self.state.rewrites:
                rw = self.state.rewrites[part[3]]
                if i not in self.graph_cache:
                    if not rw.stub:
                        if rw.status == RewriteState.UNCHECKED:
                            rw.status = RewriteState.CHECKING
                            def check_finished(i: int) -> Callable:
                                def f() -> None:
                                    self.graph_cache.pop(i, None)
                                    self.show_at_cursor()
                                return f

                            check_thread = CheckThread(rw, self)
                            check_thread.finished.connect(check_finished(i))
                            check_thread.start()

                    lhs = rw.lhs.copy()
                    rhs = rw.rhs.copy()
                    convex_layout(lhs)
                    convex_layout(rhs)
                    self.graph_cache[i] = (lhs, rhs)
                else:
                    lhs, rhs0 = self.graph_cache[i]
                    if not rhs0: raise ValueError("Rewrite step in graph_cache should have RHS")
                    rhs = rhs0

                if rw.status == RewriteState.VALID:
                    self.code_view.set_current_region((part[0], part[1]), status=STATUS_GOOD)
                elif rw.status == RewriteState.INVALID:
                    self.code_view.set_current_region((part[0], part[1]), status=STATUS_BAD)

                self.rhs_view.setVisible(True)
                self.lhs_view.set_graph(lhs)
                if rhs: self.rhs_view.set_graph(rhs)


    def next_rewrite_at_cursor(self) -> None:
        self.update_state()
        if not self.parsed: return

        pos = self.code_view.textCursor().position()
        part = self.state.part_at(pos)
        if part and part[2] == 'rewrite' and part[3] in self.state.rewrites:
            rw = self.state.rewrites[part[3]]
            start, end = rw.term_pos
            text = self.code_view.toPlainText()
            term = text[start:end]

            found_prev = (term == '?')
            rw_term = None
            for m in match_rule(rw.rule, rw.lhs):
                t = graph_to_term(rewrite(rw.rule, m))
                if found_prev and term != t:
                    rw_term = t
                    break
                elif not rw_term:
                    rw_term = t

                found_prev = (term == t)

            if rw_term:
                cursor = self.code_view.textCursor()
                cursor.clearSelection()
                cursor.setPosition(start)
                cursor.setPosition(end, mode=QTextCursor.MoveMode.KeepAnchor)
                cursor.insertText(rw_term)
                self.code_view.setTextCursor(cursor)
                self.update_state()

    def repeat_step_at_cursor(self) -> None:
        self.update_state()
        pos = self.code_view.textCursor().position()
        part = self.state.part_at(pos)
        if part and part[2] == 'rewrite' and part[3] in self.state.rewrites:
            rule = self.state.rewrites[part[3]].rule
            self.code_view.add_line_below('  = ? by ' + rule.name)
            self.update_state()
            self.next_rewrite_at_cursor()

    def update_state(self, quiet: bool=False) -> None:
        self.code_view.set_current_region(None)
        self.state.update(self.code_view.toPlainText())

        if len(self.state.errors) == 0:
            self.lhs_view.set_graph(Graph())
            self.rhs_view.setVisible(False)
            self.parsed = True
            self.show_at_cursor()
        elif not quiet:
            print('**********************************************************************')
            for err in self.state.errors:
                print("%d: %s" % err)

    def closeEvent(self, e: QCloseEvent) -> None:
        if self.doc.confirm_close():
            conf = QSettings('chyp', 'chyp')
            conf.setValue("editor_window_geometry", self.saveGeometry())
            conf.setValue("editor_splitter_state", self.splitter.saveState())
            e.accept()
        else:
            e.ignore()

class CheckThread(QThread):
    def __init__(self, rw: RewriteState, parent: Optional[QObject] = None) -> None:
        super().__init__(parent)
        self.rw = rw

    def run(self) -> None:
        self.rw.check()



#     chyp - A compositional hypergraph library
#     Copyright (C) 2022 - Aleks Kissinger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Optional, List, Tuple

from PySide6.QtCore import Qt, QPointF
from PySide6.QtGui import QBrush, QColor, QFont, QPainter, QPainterPath, QPen, QTransform, QUndoStack
from PySide6.QtWidgets import (QGraphicsEllipseItem, QGraphicsItem,
                               QGraphicsPathItem, QGraphicsRectItem, QGraphicsScene,
                               QGraphicsSceneMouseEvent, QStyleOptionGraphicsItem, QWidget)

from ..graph import Graph

SCALE = 50.0

class EItem(QGraphicsRectItem):
    def __init__(self, g: Graph, e: int) -> None:
        super().__init__(-0.4 * SCALE, -0.8 * SCALE, 0.8 * SCALE, 1.6 * SCALE)
        self.g = g
        self.e = e
        ed = g.edge_data(e)
        self.num_s = len(ed.s)
        self.num_t = len(ed.t)
        self.is_id = isinstance(ed.value, str) and str(ed.value) == 'id'

        self.setPos(ed.x * SCALE, ed.y * SCALE)

        if self.is_id:
            self.setRect(-0.2 * SCALE, -0.2 * SCALE, 0.4 * SCALE, 0.4 * SCALE)
            self.setPen(QPen(QColor(200,200,200)))
            self.setBrush(QBrush(QColor(200,200,200)))
            self.setVisible(False)
        else:
            if self.num_s <= 1 and self.num_t <= 1:
                self.setRect(-0.4 * SCALE, -0.4 * SCALE, 0.8 * SCALE, 0.8 * SCALE)

            self.setBrush(QBrush(QColor(200,200,255)))
            if ed.highlight:
                # pen = QPen(QColor(0,150,0))
                pen = self.pen()
                pen.setWidth(3)
                self.setPen(pen)
            else:
                self.setPen(QPen(QColor(0,0,0)))

    def paint(self, painter: QPainter, option: QStyleOptionGraphicsItem, widget: Optional[QWidget]=None) -> None:
        super().paint(painter, option, widget)
        ed = self.g.edge_data(self.e)

        painter.setFont(QFont("sans", 14))
        painter.drawText(self.boundingRect(), Qt.AlignmentFlag.AlignCenter, str(ed.value)) # type:ignore

class VItem(QGraphicsEllipseItem):
    def __init__(self, g: Graph, v: int) -> None:
        super().__init__(-0.0625 * SCALE, -0.0625 * SCALE, 0.125 * SCALE, 0.125 * SCALE)
        self.setVisible(False)
        self.g = g
        self.v = v
        vd = g.vertex_data(v)
        self.setPos(vd.x * SCALE, vd.y * SCALE)
        self.setBrush(QBrush(QColor(0,0,0)))

class TItem(QGraphicsPathItem):
    def __init__(self, vitem: VItem, eitem: EItem, i: int, src: bool) -> None:
        super().__init__()
        g = vitem.g
        self.vitem = vitem
        self.eitem = eitem
        if (g.vertex_data(vitem.v).highlight or
            g.edge_data(eitem.e).highlight):
            pen = QPen()
            pen.setWidth(3)
            self.setPen(pen)
        self.i = i
        self.src = src
        self.refresh()

    def refresh(self) -> None:
        path = QPainterPath()

        x_shift = (0 if self.eitem.is_id else 0.4) * SCALE

        if self.src:
            if self.eitem.num_s == 1:
                y_shift = 0.0
            else:
                y_shift = ((self.i / (self.eitem.num_s - 1)) - 0.5) * SCALE

            p1x = self.vitem.pos().x()
            p1y = self.vitem.pos().y()
            p2x = self.eitem.pos().x() - x_shift
            p2y = self.eitem.pos().y() + y_shift
        else:
            if self.eitem.num_t == 1:
                y_shift = 0
            else:
                y_shift = ((self.i / (self.eitem.num_t - 1)) - 0.5) * SCALE

            p1x = self.eitem.pos().x() + x_shift
            p1y = self.eitem.pos().y() + y_shift
            p2x = self.vitem.pos().x()
            p2y = self.vitem.pos().y()

        dx = abs(p1x - p2x)
        path.moveTo(p1x, p1y)
        path.cubicTo(p1x + dx * 0.4, p1y,
                     p2x - dx * 0.4, p2y,
                     p2x, p2y)
        self.setPath(path)
        self.update(-2000,-2000,4000,4000)

class GraphScene(QGraphicsScene):
    def __init__(self) -> None:
        super().__init__()
        self.undo_stack = QUndoStack(self)

        self.setSceneRect(-2000, -2000, 4000, 4000)
        self.setBackgroundBrush(QBrush(QColor(255,255,255)))
        self.drag_start = QPointF(0,0)
        self.drag_items: List[Tuple[QGraphicsItem, QPointF]] = []

    def set_graph(self, g: Graph) -> None:
        self.g = g
        self.clear()
        self.add_items()
        self.invalidate()

    def add_items(self) -> None:
        vi = {}
        ei = {}
        for e in self.g.edges():
            ei[e] = EItem(self.g, e)
            self.addItem(ei[e])

        for v in self.g.vertices():
            vi[v] = VItem(self.g, v)
            self.addItem(vi[v])

        for e in self.g.edges():
            ed = self.g.edge_data(e)
            for i, v in enumerate(ed.s):
                ti = TItem(vi[v], ei[e], i, src=True)
                self.addItem(ti)
            for i, v in enumerate(ed.t):
                ti = TItem(vi[v], ei[e], i, src=False)
                self.addItem(ti)

    def mousePressEvent(self, e: QGraphicsSceneMouseEvent) -> None:
        super().mousePressEvent(e)
        
        self.drag_start = e.scenePos()

        for it in self.items(e.scenePos(), deviceTransform=QTransform()):
            if it and (isinstance(it, EItem) or isinstance(it, VItem)):
                self.drag_items = [(it, it.scenePos())]
                break

    def mouseMoveEvent(self, e: QGraphicsSceneMouseEvent) -> None:
        p = e.scenePos()
        grid_size = SCALE / 8
        dx = round((p.x() - self.drag_start.x())/grid_size) * grid_size
        dy = round((p.y() - self.drag_start.y())/grid_size) * grid_size

        # move the items that have been dragged
        for it,pos in self.drag_items:
            it.setPos(QPointF(pos.x() + dx, pos.y() + dy))

        # update positions for any tentacles attached to dragged items
        for it in self.items():
            if isinstance(it, TItem):
                for it1,_ in self.drag_items:
                    if it.vitem == it1 or it.eitem == it1:
                        it.refresh()
                        break
            elif isinstance(it, VItem) or (isinstance(it, EItem) and it.is_id):
                if abs(it.pos().x() - p.x()) < 10 and abs(it.pos().y() - p.y()) < 10:
                    it.setVisible(True)
                else:
                    it.setVisible(False)


    def mouseReleaseEvent(self, _: QGraphicsSceneMouseEvent) -> None:
        self.drag_items = []




from typing import Any, Dict, List, Optional, Tuple
from lark import Lark, Transformer, UnexpectedCharacters, UnexpectedEOF, UnexpectedToken, v_args
from lark.tree import Meta

from .graph import Graph, GraphError, gen, perm, identity
from .rule import Rule, RuleError

GRAMMAR = Lark("""
    start : statement*
    ?statement : gen | let | rule | rewrite
    gen : "gen" var ":" num "->" num
    let : "let" var "=" term
    rule : "rule" var ":" term "=" term
    rewrite : "rewrite" var ":" term rewrite_part*
    rewrite_part : "=" term_hole "by" rule_ref num?
    ?term  : par_term | seq
    ?par_term : "(" term ")" | par | perm | id | term_ref
    par : par_term "*" par_term
    seq : term ";" term
    perm : "sw" [ "[" num ("," num)* "]" ]
    id : "id"

    num : INT
    var : CNAME
    term_ref : CNAME
    rule_ref : CNAME
    term_hole : term | "?"

    %import common.CNAME
    %import common.INT
    %ignore " "
    %ignore "\\n"
    """,
    parser='lalr',
    propagate_positions=True)

class ChypParseData(Transformer):

    def __init__(self) -> None:
        self.graphs: Dict[str, Graph] = dict()
        self.rules: Dict[str, Rule] = dict()
        self.rewrites: Dict[str, Tuple[int, int, Optional[Rule], Optional[Graph], Optional[Graph]]] = dict()
        self.errors: List[Tuple[int, str]] = list()
        self.parts: List[Tuple[int, int, str, str]] = list()
        self.parsed = False
    
    # def start(self, items: List[List[Tuple[int,int,str,str]]]):
    #     pass
        
    def var(self, items: List[Any]) -> str:
        return str(items[0])
    
    def num(self, items: List[Any]) -> int:
        return int(items[0])

    def id(self, _: List[Any]) -> Graph:
        return identity()

    @v_args(meta=True)
    def perm(self, meta: Meta, items: List[Any]) -> Optional[Graph]:
        try:
            if items[0] is None:
                return perm([1,0])
            else:
                return perm([int(i) for i in items])
        except GraphError as e:
            self.errors.append((meta.line, str(e)))
        return None
    
    @v_args(meta=True)
    def term_ref(self, meta: Meta, items: List[Any]) -> Optional[Graph]:
        s = str(items[0])
        if s in self.graphs:
            return self.graphs[str(items[0])]
        else:
            self.errors.append((meta.line, 'Undefined term: ' + s))
            return None

    @v_args(meta=True)
    def rule_ref(self, meta: Meta, items: List[Any]) -> Optional[Rule]:
        s = str(items[0])
        if s in self.rules:
            return self.rules[str(items[0])]
        else:
            self.errors.append((meta.line, 'Undefined term: ' + s))
            return None
    
    def par(self, items: List[Any]) -> Optional[Graph]:
        if items[0] and items[1]:
            return items[0] * items[1]
        else:
            return None
    
    @v_args(meta=True)
    def seq(self, meta: Meta, items: List[Any]) -> Optional[Graph]:
        if items[0] and items[1]:
            g = None
            try:
                g = items[0] >> items[1]
            except GraphError as e:
                self.errors.append((meta.line, str(e)))
            return g
        else:
            return None

    @v_args(meta=True)
    def gen(self, meta: Meta, items: List[Any]) -> None:
        name, arity, coarity = items
        self.graphs[name] = gen(name, arity, coarity)
        self.parts.append((meta.start_pos, meta.end_pos, 'gen', name))
        
    @v_args(meta=True)
    def let(self, meta: Meta, items: List[Any]) -> None:
        name, graph = items
        if graph:
            self.graphs[name] = graph
        self.parts.append((meta.start_pos, meta.end_pos, 'let', name))

    @v_args(meta=True)
    def rule(self, meta: Meta, items: List[Any]) -> None:
        name, lhs, rhs = items
        if lhs and rhs:
            try:
                self.rules[name] = Rule(lhs, rhs, name=name)
            except RuleError as e:
                self.errors.append((meta.line, str(e)))
        self.parts.append((meta.start_pos, meta.end_pos, 'rule', name))

    @v_args(meta=True)
    def rewrite(self, meta: Meta, items: List[Any]) -> None:
        name = items[0]
        term = items[1]
        rw_parts = items[2:]

        if len(rw_parts) == 0:
            self.parts.append((meta.start_pos, meta.end_pos, "rewrite", name))
            self.rewrites[name] = (0,0,None,term,None)
        else:
            start = meta.start_pos
            lhs = term
            rhs = None
            for i, rw_part in enumerate(rw_parts):
                end, t_start, t_end, rule, rhs = rw_part
                self.rewrites[name + ":" + str(i)] = (t_start, t_end, rule, lhs, rhs)
                lhs = rhs.copy() if rhs else None
                self.parts.append((start, end, "rewrite", name + ":" + str(i)))
                start = end
            if rhs:
                self.rules[name] = Rule(term, rhs, name)

    @v_args(meta=True)
    def rewrite_part(self, meta: Meta, items: List[Any]) -> Tuple[int, int, int, Rule, Graph]:
        t_start,t_end,rhs = items[0]
        rule = items[1]
        return (meta.end_pos, t_start, t_end, rule, rhs)


    @v_args(meta=True)
    def term_hole(self, meta: Meta, items: List[Any]) -> Tuple[int, int, Optional[Graph]]:
        t = items[0] if len(items) != 0 else None
        return (meta.start_pos, meta.end_pos, t)

def parse(code: str) -> ChypParseData:
    parse_data = ChypParseData()
    try:
        tree = GRAMMAR.parse(code)
        parse_data.transform(tree)
        parse_data.parsed = True
    except UnexpectedEOF as e:
        parse_data.errors += [(-1, str(e))]
    except UnexpectedToken as e:
        parse_data.errors += [(e.token.line, str(e))]
    except UnexpectedCharacters as e:
        parse_data.errors += [(e.line, str(e))]

    return parse_data




from typing import List, Mapping, Tuple, Set
from .graph import Graph

# def ready_edges(g: Graph, edges: Set[int], v_pos: Mapping[int,float]) -> Set[int]:
#     ready = set()
#     for e in edges:
#         src = g.source(e)
#         if all(v in v_pos for v in src):
#             ready.add(e)
#     return ready

def layer_decomp(g: Graph) -> Tuple[List[List[int]], List[List[int]]]:
    """Decompose a graph into regular and singular layers

    Returns a pair of lists, sorting the vertices and edges of
    `g` into layers, respectively. Note that this can modify `g`
    by introducing extra vertices and identity boxes.
    """

    v_layers = []
    e_layers = []

    v_pos = dict()
    e_pos = dict()
    v_layer = []
    for i,v in enumerate(g.inputs()):
        v_layer.append(v)
        v_pos[v] = i
    edges = set(g.edges())
    new_ids = set()
    
    while len(edges) > 0:
        v_layers.append(v_layer)

        ready = set()
        for e in edges:
            if all(v in v_pos for v in g.source(e)):
                ready.add(e)
        outputs = set(g.outputs())
        for v in v_layer:
            if v in outputs or any(e not in ready for e in g.out_edges(v)):
                id = g.insert_id_after(v)
                new_ids.add(id)
                ready.add(id)
        
        e_layer = []
        for e in ready:
            src = g.source(e)
            e_pos[e] = sum(v_pos[v] for v in src)/len(src) if len(src) != 0 else 0
            e_layer.append(e)
            edges.discard(e)
        if all(e in new_ids for e in e_layer):
            raise ValueError("Could not make progress. Is graph acyclic?")

        list.sort(e_layer, key=lambda e: e_pos[e])

        e_layers.append(e_layer)

        if len(edges) > 0:
            v_layer = []
            for e in e_layer:
                for v in g.target(e):
                    v_pos[v] = len(v_layer)
                    v_layer.append(v)


    v_layers.append(g.outputs().copy())
    return (v_layers, e_layers)

def perm_to_s(perm: List[int]) -> str:
    if len(perm) == 1:
        return 'id'
    elif len(perm) == 2:
        return 'sw'
    else:
        return 'sw' + str(perm)

def split_perm(perm: List[int]) -> List[List[int]]:
    """Split a permutation into a tensor product of independent permuations
    """
    perms = []
    rest = perm

    while rest != []:
        m = 0
        for i, x in enumerate(rest):
            m = max(x, m)
            if m <= i:
                perms.append(rest[:i+1])
                rest = [y-(i+1) for y in rest[i+1:]]
                break
    return perms


def graph_to_term(g: Graph) -> str:
    """Convert a graph to a term

    This currently only works for monogamous acyclic graphs (and hence symmetric monoidal
    terms).
    """

    g = g.copy()
    v_layers, e_layers = layer_decomp(g)

    # build a list of terms to be sequentially composed
    seq = []
    for i in range(len(e_layers)):
        # compute the permulation from the current vertex layer to inputs of the edge layer
        s_layer = [v for e in e_layers[i] for v in g.source(e)]
        v_pos = { v : j for j,v in enumerate(v_layers[i]) }
        v_perm = [v_pos[v] for v in s_layer]

        # append it as a layer of swap maps
        if v_perm != list(range(len(v_perm))):
            perms = split_perm(v_perm)
            seq.append(' * '.join([perm_to_s(p) for p in perms]))

        # append the parallel composition of the current edge layer
        par = [str(g.edge_data(e).value) for e in e_layers[i]]
        seq.append(' * '.join(par))

        # compute the permulation from the outputs of the edge layer to the next vertex layer
        t_layer = [v for e in e_layers[i] for v in g.target(e)]
        v_pos = { v : j for j,v in enumerate(t_layer) }
        v_perm = [v_pos[v] for v in v_layers[i+1]]

        # append it as a layer of swap maps
        if v_perm != list(range(len(v_perm))):
            perms = split_perm(v_perm)
            seq.append(' * '.join([perm_to_s(p) for p in perms]))

    return ' ; '.join(seq)




#     chyp - A compositional hypergraph library
#     Copyright (C) 2022 - Aleks Kissinger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
import sys

from PySide6.QtWidgets import QApplication

from . import editor


class Chyp(QApplication):
    """The main chyp application

    ...
    """

    def __init__(self) -> None:
        super().__init__(sys.argv)
        self.setApplicationName('chyp')
        self.setDesktopFileName("chyp")
        self.main_window = editor.Editor()
        self.lastWindowClosed.connect(self.quit)


def main() -> None:
    """Main entry point for chyp"""

    chyp = Chyp()
    chyp.exec_()
